#include <iostream>
#include <errlog.h>
#include <astrotime.h>
#include "sub_main.h"
#include "deftypes.h"
#include "actions.h"

const std::string park_phase = "Parking: ";
const std::string init_phase = "Initialization: ";
const std::string sect_comp = "Components";
const std::string sect_gen  = "General";
const unsigned int short_delay = 3;

AstroDateTime uta;

Druid::Druid( Config& cfg ) : background( cfg, *this ), mass_enable(true) {
    settle_time = cfg.inp<unsigned int>( sect_comp, "RCX400", "GuideSettleTime" );
    separation = cfg.inp<double>( sect_comp, "DIMM", "Separation" );
    sep_tolrn  = cfg.inp<double>( sect_comp, "DIMM", "SepTolerance" );
    work_zone =  cfg.inp<double>( sect_comp, "RCX400", "WorkZone" );
    mass_mag_ext = cfg.inp<double>( sect_comp, "MASS", "MaxExtinct" );

    ClientSocket::AddrType dimmIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "DIMM", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "DIMM", "ThePort") );
    ClientSocket::AddrType massIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "MASS", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "MASS", "ThePort") );
    ClientSocket::AddrType tlspIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "RCX400", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "RCX400", "ThePort") );
    try {
        dimm = std::auto_ptr<ClientHandler>( new ClientHandler( dimmIP ) );
        mass = std::auto_ptr<ClientHandler>( new ClientHandler( massIP ) );
        tlsp = std::auto_ptr<ClientHandler>( new ClientHandler( tlspIP ) );
    }
    catch(const std::exception& e ) {
        ErrlogSingl::instance().write( std::string("ASM components: ") + e.what() );
        throw ErrFat( ERR_DEV_FUNC, "ASM components connection fails" );
    }
    try {
        if( dimm->wait_nonbusy( "init" ) && dimm->status( "READY" ) )
            ErrlogSingl::instance().write( init_phase + "DIMM ready" );
        if( mass->wait_nonbusy( "init" ) && mass->status( "READY" ) )
            ErrlogSingl::instance().write( init_phase + "MASS ready" );
        if( tlsp->wait_nonbusy( "init" ) && tlsp->status( "READY" ) )
            ErrlogSingl::instance().write( init_phase + "RCX400 ready" );
        if( tlsp->status( "ERFAT" ) ) throw std::runtime_error(" Server fatal error" );
    }
    catch(const std::exception& e ) {
        sleep(5*short_delay);
        if( dimm->get_status( "READY" ) ) dimm->wait_nonbusy( "park" );
        if( mass->get_status( "READY" ) ) mass->wait_nonbusy( "park" );
        if( tlsp->get_status( "READY" ) ) tlsp->wait_nonbusy( "park" );
        ErrlogSingl::instance().write( std::string("Initialization: ") + e.what() );
        throw ErrFat( ERR_DEV_FUNC, "ASM components initialization fails" );
    }
}

Druid::~Druid() {
    try {
        background.run_nochecked();
        stop_guiding();
        sleep(short_delay);
    }
    catch(const std::exception& e ) {
        ErrlogSingl::instance().write( std::string("Finalize work: ") + e.what() );
    }
    if( park( dimm.get() ) ) ErrlogSingl::instance().write( park_phase + "DIMM stopped" );
    if( park( mass.get() ) ) ErrlogSingl::instance().write( park_phase + "MASS stopped" );
    if( park( tlsp.get() ) ) ErrlogSingl::instance().write( park_phase + "RCX400 stopped" );
//    std::cerr << "Druid  destructor" << std::endl;
}

bool Druid::park( ClientHandler* comp ) {
    try {
        if( comp->status( "BUSY" ) ) {
            comp->command( "stop" );
            sleep( 2*short_delay );
        }
        if( comp->wait_nonbusy( "park" ) && comp->status( "PARKED" ) ) return true;
    }
    catch(const std::exception& e ) {
        ErrlogSingl::instance().write( std::string("Park Druid components: ") + e.what() );
    }
    return false;
}

bool Druid::point_star( const ObjectInfo& obj ) try {
    int acount = 0;
    stop_guiding();
    if( !background.no_needed() ) background.run_checked();               // Background measurement before repointing
    ErrlogSingl::instance().write( "Start point to " + ::Name( obj ) );
    while( acount < max_retries ) {
        if( acount > 0 ) {
            ErrlogSingl::instance().write( ::Name( obj ) + " is not found. Will try again" );
            sleep( settle_time/2 );
        }
        tlsp->command( "run ra=\"" + ::RA( obj ) + "\" dec=\"" + ::DEC( obj ) +"\"" );
        if( tlsp->status( "READY" ) ) {
            ErrlogSingl::instance().write( ::Name( obj ) + " is found" );
            dimm->wait_nonbusy( "set object=\"" + OBJECT( obj ) + "\"" );
            mass->wait_nonbusy( "set object=\"" + OBJECT( obj ) + "\"" );
            background.run_checked();       // Background measurement after pointing
            return center_star( obj );
        }
        acount++;
    }
    ErrlogSingl::instance().write( ::Name( obj ) + " is not found! Ask eigther star" );
    return false;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Pointing: ") + e.what() );
    return false;
}

bool Druid::center_star( const ObjectInfo& obj ) try {
    int acount = 0;
    while( acount < max_retries ) {
        if( acount > 0 ) {
            ErrlogSingl::instance().write( "Star centering failed. Will try again" );
            sleep(settle_time/4);
        }
        tlsp->command( "run center" );
        if( tlsp->status( "READY" ) && start_guiding() ) {
            sleep( settle_time/2 );
            if( tlsp->command( "get offset" ) ) {
                point<double>offs = point<double>( tlsp->get_value<double>( "X" ), tlsp->get_value<double>( "Y" ) );
                if( offs.distance() < work_zone ) {
                    ErrlogSingl::instance().write(  ::Name( obj ) + " is centered" );
                    return true;            // Test of guiding
                }
            }
        }
        acount++;
    }
    ErrlogSingl::instance().write( "Star centering failed" );
    return false;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Centering: ") + e.what() );
    return false;
}

bool Druid::check_star( const ObjectInfo& obj ) try {
    errflux = 1;
    if( mass_enable ) {
        if( !check_flux() ) {
            center_star( obj );
            check_flux();
        }
        if( !check_mass_mag( obj, mass_mag_ext ) ) return false;
    }
    stop_guiding();
    start_dimm_guiding();
    dimm->command( "get offset" );                      //  
    point<double>offs = point<double>( dimm->get_value<double>( "X" ), dimm->get_value<double>( "Y" ) );
    if( offs.distance() < work_zone/5 ) {
        ErrlogSingl::instance().write( "DIMM guiding works" );
    } else {}       //??
    tlsp->command( "set sync=on" );           //     
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Checking: ") + e.what() );
    return false;
}

bool Druid::measurement() try {
    int max_att = max_attempts;
    dimm->nonblocked( "run" );
    while( dimm->status( "BUSY" ) ) {
        if(max_att-- < 0 ) throw std::runtime_error("DIMM very busy" );
        usleep( 500000 );
        dimm->nonblocked( "run" );
    }
    mass->nonblocked( "run" );
    dimm->wait_command();
    mass->wait_command();
    if( !mass->status( "READY" ) || !dimm->status( "READY" ) ) errcount++;
    if( mass->command( "get flux" ) ) star_flux_d = mass->get_value<double>( "Flux_D" );
    if( star_flux_d < 3*background.get() ) return false;
    mass_mag_set.push_back( star_flux_d );
    return true;
}
catch(const ErrFat& e ) {
    errcount++;
    if( e.err() == ERR_SOCK_BAD ) throw ErrFat( e );
    return false;
}
catch(const std::exception& e ) {
    errcount++;
    ErrlogSingl::instance().write( std::string("Measurement: ") + e.what() );
    return false;
}

bool Druid::dimm_measurement() try {
    int max_att = 8;
    mass->command( "set hv=off" );
    dimm->nonblocked( "run" );
    while( dimm->status( "BUSY" ) ) {
        if(max_att-- < 0 ) throw std::runtime_error("DIMM very busy" );
        usleep( 500000 );
        dimm->nonblocked( "run" );
    }
    dimm->wait_command();
    if( !dimm->status( "READY" ) ) errcount++;
    return true;
}
catch(const ErrFat& e ) {
    errcount++;
    if( e.err() == ERR_SOCK_BAD ) throw ErrFat( e );
    return false;
}
catch(const std::exception& e ) {
    errcount++;
    ErrlogSingl::instance().write( std::string("DIMM Measurement: ") + e.what() );
    return false;
}

bool Druid::beginning() try {
    errcount = 0;
    mass_mag_set.clear();
    return focus_dimm();
}
catch(const ErrFat& e ) {
    if( e.err() == ERR_SOCK_BAD ) throw ErrFat( e );
    return false;
}
catch(const std::exception& e ) {
    errcount++;
    ErrlogSingl::instance().write( std::string("Measurement: ") + e.what() );
    return false;
}

bool Druid::check_status( const ObjectInfo& obj ) try {
    message<int,0>( "Measurement set is completed. Errors=", errcount );
    if( background.needed() && background.run_checked() ) {
        tlsp->command( "run center" );
        start_dimm_guiding();
    }
    if( !check_mass_mag( obj, mass_mag_ext ) ) return false;
    ErrlogSingl::instance().write( "Checking successful" );
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Checking: ") + e.what() );
    return false;
}

bool Druid::no_check_status( const ObjectInfo& obj ) try {
    message<int,0>( "Experimental set is completed. Errors=", errcount );
    if( background.run_nochecked() ) {
        tlsp->command( "run center" );
        start_dimm_guiding();
    }
//    check_mass_mag( obj, mass_mag_ext );
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Experiment checking: ") + e.what() );
    return false;
}

bool Druid::check_background( const ObjectInfo& obj ) try {
    if( background.run_nochecked() ) {
        tlsp->command( "run center" );
        start_dimm_guiding();
    }
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("Experiment checking: ") + e.what() );
    return false;
}

void Druid::regime( const std::string& s ) {
    static std::string prevs = "normal";
    if( s != prevs ) {
        if( s == "normal" ) {
            camera("high");
            mass_enable = true;
        }
        if( s == "massdimm" ) {
            camera("medium");
            mass_enable = true;
        }
        if( s == "dimm" ) {
            camera("low");
            mass_enable = false;
        }
        prevs = s;
    }
}

void Druid::camera( const std::string& s ) {
    static std::string prevs = "high";
    if( s != prevs ) {
        tlsp->command( "set camera="+s );
        prevs = s;
    }
}

void Druid::focusing( double f ) {
    if( fabs( f ) < 0.002 ) return;
    f = sign(f)*min( fabs(f), 0.05 );
    focus += f;
    std::ostringstream ostr;
    ostr << "run focus=" << std::fixed << std::setprecision(3) << f;
    tlsp->command( ostr.str() );
}

double Druid::check_separation() {
    dimm->command( "run center" );
    if( dimm->status( "ERFAT" ) ) return 0.0;
    if( dimm->status( "BUSY" ) ) sleep(2);
    dimm->command( "get images" );
    if( dimm->get_value<double>( "dX" )==0 && dimm->get_value<double>( "dY" )==0 ) return 0.0;
    return separation - dimm->get_value<double>( "dX" );
}

bool Druid::focus_dimm() {
    const int ittmax = 2;
    int icount = 0;
    double dS = check_separation();
    while( fabs(dS) > sep_tolrn && icount++ < ittmax ) {
        focusing( 0.004*dS );
        dS = check_separation();
    }
    if( uta.currentSec() - guide_last > 1000.0 ) {
        guide_last = uta.currentSec();
        tlsp->command( "get center" );
        message<int,0>( "Optical center X = ", tlsp->get_value<double>( "X" ), " Y = ", tlsp->get_value<double>( "Y" ) );
    }
    return fabs(dS) < sep_tolrn;
}

bool Druid::check_flux() {
    mass->command( "run flux" );
    if( mass->command( "get flux" ) ) star_flux_d = mass->get_value<double>( "Flux_D" );
    if( star_flux_d < 3*background.get() ) return false;
    return true;
}

bool Druid::photo_stability(){
    if( mass_mag_set.size() > 2 ) {
        double minflux = *std::min_element(mass_mag_set.begin(), mass_mag_set.end());
        double maxflux = *std::max_element(mass_mag_set.begin(), mass_mag_set.end());
        if( maxflux - minflux <  0.02*(maxflux + minflux) ) return true;
    }
    return false;
}

bool Druid::check_mass_mag( const ObjectInfo& obj, double mass_mag_ext ) {
    double mass_cnt = MASScount( obj );
//    if( mass->command( "get flux" ) ) star_flux_d = mass->get_value<double>( "Flux_D" );
    message<double,0>( "Calculated MASS Flux_D: ", 0.794*mass_cnt, ", measured: ", star_flux_d );
    if( star_flux_d < mass_mag_ext*mass_cnt || star_flux_d > mass_cnt/mass_mag_ext  ) {
        ErrlogSingl::instance().write( "Low (incorrect) star fluxes" );
        if( errflux++ > 0 ) return false;
    } else {
        errflux = 0;
    }
    return true;
}

bool Druid::start_guiding() {
    tlsp->wait_nonbusy( "set guide=on" );
    if( tlsp->status( "READY" ) ) { sleep(short_delay); return true; }
    ErrlogSingl::instance().write( "Guiding start failed" );
    return false;
}

bool Druid::start_dimm_guiding() {
    tlsp->wait_nonbusy( "set guide=dimm" );
    if( tlsp->status( "READY" ) ) {
        sleep( settle_time );
        guide_last = uta.currentSec();
        return true;
    }
    ErrlogSingl::instance().write( "DIMM guiding start failed" );
    return false;
}

bool Druid::stop_guiding() {
    tlsp->command( "get guide" );
    if( tlsp->get_value<std::string>( "GUIDE" ) == "off" ) return true;
    tlsp->wait_nonbusy( "set guide=off" );
    if( tlsp->status( "READY" ) ) { sleep(short_delay); return true; }
    ErrlogSingl::instance().write( "Guiding stop failed" );
    return false;
}

Background::Background( Config& cfg, Druid& ac ) : act(ac), value(-1), errcount(0) {
    displace = cfg.inp<double>( sect_gen, "Background", "Displacement" );
    period   = cfg.inp<double>( sect_gen, "Background", "Period" );
    same     = cfg.inp<double>( sect_gen, "Background", "MinimalTime" );
    maximal  = cfg.inp<double>( sect_gen, "Background", "MaxValue" );
    last = uta.currentSec();
}

bool Background::needed() {
    return ( uta.currentSec() < last + period )? false : true;
}

bool Background::no_needed() {
    return ( uta.currentSec() < last + same )? true : false;
}

bool Background::run_checked() {
    int acount = 0;
    if( !act.mass_enable ) return false;
    act.stop_guiding();
    while( acount++ < max_count ) {
        double curr = measure();
        if( curr < maximal ) {
            value = curr;
            message<double,2>( "Background in D aperture: ", value );
            return true;
        }
        ErrlogSingl::instance().write( "Extremal background, Will try again" );
        sleep(short_delay);
    }
    ErrlogSingl::instance().write( "Background estimation problem" );
    return false;
}

bool Background::run_nochecked() {
    if( !act.mass_enable ) return false;
    act.stop_guiding();
    value = measure();
    message<double,2>( "Background in D aperture: ", value );
    return true;
}

double Background::measure() {
    double f = -1.0;
    point<double> offs;
    if( act.tlsp->command( "get offset" ) ) {
        offs = point<double>( act.tlsp->get_value<double>( "X" ), act.tlsp->get_value<double>( "Y" ) );
        if( offs.distance()  < 10.0 ) offs = point<double>( 10.0, 10.0 );
        f = displace/offs.distance() - 1.0;
        if( f > 0 ) {
            std::ostringstream ostr;
            f = 1.3*f + 0.3;
            ostr << std::fixed << std::setprecision(0) << "run dx=" << -f*offs.x() << " dy=" << -f*offs.y();
            act.tlsp->command( ostr.str() );
        }
    }
    act.mass->command( "run background" );
    if( act.mass->status( "READY" ) ) {
        act.mass->command( "get flux" );
        last = uta.currentSec();
        return act.mass->get_value<double>("Flux_D");
    }
    return 0.0;
}
