#include <iostream>
#include <errlog.h>
#include <astrotime.h>

#include "deftypes.h"
#include "omicron.h"

const std::string start_phase = "Power on: ";
const std::string sect_comp = "Components";
const unsigned int short_delay = 3;

AstroDateTime ut;

Omicron::Omicron( Config& cfg, bool m ) try : man_mode(m) {
    delay_to_druid = cfg.inp<unsigned int>( sect_comp, "Dome", "StartDelay");
    delay_from_druid = cfg.inp<unsigned int>( sect_comp, "Dome", "HaltDelay");
    ClientSocket::AddrType domeIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "Dome", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "Dome", "ThePort") );
    dome = std::auto_ptr<ClientHandler>( new ClientHandler( domeIP ) );
    shutIP = IP4Addr( cfg.inp<std::string>( sect_comp,  "Druid", "TheHost"),
                                             cfg.inp<unsigned int>( sect_comp, "Druid", "ThePort") );
    host_checked = cfg.inp<std::string>( sect_comp, "Druid", "TheHost");
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("ASM start: ") + e.what() );
    throw ErrFat( ERR_DEV_FUNC, "ASM connection fails" );
}

Omicron::~Omicron() {
    char ans = 'Y';
    if( man_mode ) {    //  !!!
        std::cout << " ***** Power off instruments and close dome? " << std::flush;
        std::cin >> ans;
    }
    if( toupper(ans) == 'Y' ) finish();
//    std::cerr << "Omicron destructor" << std::endl;
}

bool Omicron::testDruid() {
    std::string req = "ping -c1 " + host_checked + " > /dev/null";
    return system( req.c_str() ) == 0;
}

bool Omicron::haltDruid() try {
    std::auto_ptr<ClientHandler> shut_daem = std::auto_ptr<ClientHandler>( new ClientHandler( shutIP ) );
    shut_daem->command( "get lock" );
    if( shut_daem->get_value<int>( "lock" ) > 0 ) {
        ErrlogSingl::instance().write( "Druid shutdown is externally locked" );
        return  false;
    }
    shut_daem->command( "run shutdown" );
    if( !shut_daem->status( "ERFAT" ) ) {
        ErrlogSingl::instance().write( "Druid will shutdown in 1 minute" );
        return true;
    }
    return false;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("ASM finish: ") + e.what() );
    return false;
}

// port 16100
// 01 run shutdown
// 01 ERROR STATUS=ERFAT LOCK=1
// 02 set unlock=abcd
// 02 OK LOCK=0
// 01 run shutdown
// 01 OK
// 02 run cancel
// 02 OK
// 03 get lock
// 03 OK LOCK=0
// 04 set lock=qwerty
// 04 OK LOCK=1
// 05 set unlock=qwerty
// 05 OK LOCK=0

bool Omicron::start() try {
    double fin_time = -1.0; 
                                //        !!!
    while( !dome->get_status( "READY" ) ) {
        if( dome->status( "ERFAT" ) ) throw ErrFat( ERR_DEV_FUNC, "Dome component fails" );
        sleep( 4*short_delay );
    }
    ErrlogSingl::instance().write( "Dome component is ready" );
    dome->command( "get power" );
    if( dome->get_value<std::string>( "power" ) == "ON" && !testDruid() ) {
        dome->command( "set power=off" );
        sleep( 4*short_delay );
        dome->command( "get power" );
    }
    if( dome->get_value<std::string>( "power" ) == "OFF" ) {
        dome->command( "set power=on" );
        fin_time = ut.currentSec() + delay_to_druid;
    }
    dome->command( "get dome" );
    if( dome->get_value<std::string>( "dome" ) != "OPENED" ) {
        dome->command( "run dome=open" );
        dome->command( "get dome" );
    }
//    if( dome->get_value<std::string>( "dome" ) == "UNDEFINED" ) ErrlogSingl::instance().write( "Dome is opened" );
    if( dome->get_value<std::string>( "dome" ) == "OPENED" ) ErrlogSingl::instance().write( "Dome is opened" );
    else throw ErrFat( ERR_DEV_FUNC, "Dome opening fails" );

    if( fin_time > 0.0 ) {
        std::ostringstream ostr;
        ostr << (int)(fin_time - ut.currentSec()) << " sec";
        ErrlogSingl::instance().write( "Wait for druid start " + ostr.str() );
        while( ut.currentSec() < fin_time ) sleep( 1 );
    }
    if( !testDruid() ) throw ErrFat( ERR_DEV_FUNC, "Druid start fails" );
    ErrlogSingl::instance().write( "Druid is started" );
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("ASM start: ") + e.what() );
    dome->command( "run dome=close" );
    dome->command( "set power=off" );
    throw ErrFat( ERR_DEV_FUNC, "ASM start fails" );
    return false;
}

bool Omicron::finish() try {
    double fin_time = -1.0;
    ErrlogSingl::instance().write( "Finish of the observation run" );
    fin_time = ut.currentSec() + delay_from_druid;
    bool unlock = haltDruid();
    dome->command( "run dome=close" );
    if( unlock && fin_time > 0.0 ) {
        std::ostringstream ostr;
        ostr << (int)(fin_time - ut.currentSec()) << " sec";
        ErrlogSingl::instance().write( "Wait druid stop " + ostr.str() );
        while( ut.currentSec() < fin_time ) sleep( 1 );
    }
    if( unlock ) dome->command( "set power=off" );
    sleep(12);
    dome->command( "get power" );
    if( dome->get_value<std::string>( "power" ) == "OFF" ) ErrlogSingl::instance().write( "ASM power is off!" );
    sleep(short_delay);
    return true;
}
catch(const std::exception& e ) {
    ErrlogSingl::instance().write( std::string("ASM finish: ") + e.what() );
    return false;
}




